/*
 * Speak Recorder
 * Copyright  2003 by Schou Industries ApS.
 */

#ifdef HAS_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <termios.h>

#include "speakrec.h"
#include "tapebox.h"

#define MASK_OFF_USED_KEYS 0xfe

unsigned char negateinput = 0x00;

struct ssegment {
	char chr;
	char *pattern;
};

struct ssegment segmenttext[] = {
#include "segment7.inc"
};

#define SIZEOF_SEGMENTTEXT sizeof(segmenttext) / sizeof(struct ssegment)

unsigned char segmentbin[128];

int lookupchr( char c ) {
	int n = 0;

	c = toupper(c);
	while (n < SIZEOF_SEGMENTTEXT && c != segmenttext[n].chr)
		n++;
	return n >= SIZEOF_SEGMENTTEXT ? 0 : n;
}

/*
 * Mapping of 7-segment display.
 * On the left side of the drawing below,
 * write the letter which match the code on the display.
 * In the sympos array, write the number from the right
 * side which matches the letter on the left side.
 *
 * Q1 E
 * Q2 G
 * Q3 D
 * Q4 F
 * Q5 A
 * Q6 C
 * Q7 B
 * Q8 dp
 *
 * .A. ._. 012
 * FGB |_| 345
 * EDC |_| 678
 */

/* Q number:           1 2 3 4 5 6 7
 * Letter left side:   E G D F A C B
const char sympos[] = {6,4,7,3,1,8,5};
*/

/* Letter left side:   A B C D E F G
 * Bit number:         0 1 2 3 4 5 6 */
const char sympos[] = {1,5,8,7,6,3,4};

int texttobin( char c ) {
	int i = 0;
	int bin = 0;
	char *pattern = segmenttext[lookupchr(c)].pattern;
	while (i<sizeof(sympos)) {
		if (pattern[sympos[i]] != ' ')
			bin |= 1 << i;
		i++;
	}
	return bin;
}

int maptexttobin( void ) {
	int i = 0;
	int bin = texttobin('?');
	char c;

	while (i<sizeof(segmentbin))
		segmentbin[i++] = bin;
	i = 0;
	while (i<SIZEOF_SEGMENTTEXT) {
		c = segmenttext[i].chr;
		segmentbin[c] = texttobin(c);
		i++;
	}

	return 0;
}

struct {
	unsigned char key_code;
	unsigned char led_code;
	tape_key key;
	button_name but_name;
} tbkeys[] = {		/* [....] Turn on ekstra bit on play or edit */
	{ 1 << 7, 1 << 0 | 1 << 7, key_edit, but_edit },
	{ 1 << 6, 1 << 1 | 0 << 7, key_record, but_record },
	{ 1 << 5, 1 << 2 | 0 << 7, key_stop, but_stop },
	{ 1 << 4, 1 << 3 | 0 << 7, key_track_next, but_track_next },
	{ 1 << 3, 1 << 4 | 0 << 7, key_track_prev, but_track_prev },
	{ 1 << 2, 1 << 5 | 0 << 7, key_menu, but_menu },
	{ 1 << 1, 1 << 6 | 1 << 7, key_play, but_play },
	{ 1 << 0, 1 << 7 | 0 << 7, key_none, but_none }
};

#define SIZE_TBKEYS (sizeof(tbkeys) / sizeof(tbkeys[0]))

#define CONSOLE_NAME "Tapebox"

/* Hardware addresses as wired */
#define IO_INPTS "A"
#define IO_ADDRS "B"
#define IO_DATAS "C"

#define IO_INPTC IO_INPTS[0]
#define IO_DATAC IO_DATAS[0]
#define IO_ADDRC IO_ADDRS[0]

#define PORT_OUTPUT "\x00"
#define PORT_INPUT  "\xFF"

#define IO_INIT "!" IO_ADDRS PORT_OUTPUT "!" IO_DATAS PORT_OUTPUT "!" IO_INPTS PORT_INPUT "2"

/* Address setup for output */
#define IO_SEGMENT 0x00
#define IO_VOLUME_LOW 0x07
#define IO_VOLUME_HIGH 0x08
#define IO_LED 0x09
#define IO_UNDEF 0x0F

/*
 * Address layout for segments and led
 *            dcba #
 * Fil x 10   0000 0
 * Fil x  1   0001 1
 * Min x 10   0010 2
 * Min x  1   0011 3
 * Sec x 10   0100 4
 * Sec x  1   0101 5
 * Sec x  0.1 0110 6
 * Ned VU     0111 7
 * Op  VU     1000 8
 * LED        1001 9
 * undef      1010 A
 * undef      1011 B
 * undef      1100 C
 * undef      1101 D
 * undef      1110 E
 * Reset      1111 F
 *
 */

int TapeboxConsole::senddata(unsigned char addr, unsigned char value) {
	int num_written = 0;
	unsigned char temp[] = {
		IO_DATAC, ~value,
		IO_ADDRC, addr,
		IO_ADDRC, IO_UNDEF
	};

	ondisplay[addr & 0x0F] = value;
	if (fd) {
		if (num_written = write(fd, &temp, sizeof(temp)) != sizeof(temp)) {
			fd = 0;
		}
	} else {
		reconnect();
	}

	return num_written;
}

int TapeboxConsole::flush(void) {
	/* if data in out-buffer, send it */
}

int TapeboxConsole::reconnect() {
	char temp[100];
	int num_read;
	fd_set read_set;
	struct timeval tv;
	int retval;
	struct termios term;

	prev_key = key_none;
	if (0 > (fd = open(devicename, O_RDWR | O_NOCTTY ))) {
		/* OK, we try later */
		fd = 0;
		fprintf(stderr, "Could not open device '%s'. Error: %s\n", devicename, strerror(errno));
	} else {
		/* Delay for open device. It take some time. */
		sleep(1);
		/* Setup: Do not act on ISIG (Ctrl-C) and so on */
		ioctl(fd, TCGETS, &term);
		term_orig = term;
		term.c_iflag &= ~( ICRNL );
		term.c_oflag &= ~( ONLCR );
		term.c_lflag &= ~( ICANON | ISIG | ECHO );
		ioctl(fd, TCSETS, &term);
		/* Identify Device */
		FD_ZERO(&read_set);
		FD_SET(fd, &read_set);
		if (write(fd, "?", 1)) {
			strncpy(temp, (char *)"*no respond from device*", sizeof(temp)-1);
			tv.tv_sec = 3;
			tv.tv_usec = 0;
			/* wait some sec for data, or give up */
			if (retval = select(fd+1, &read_set, NULL, NULL, &tv) && FD_ISSET(fd, &read_set)) {
				/* data ready for read */
				num_read = read(fd, &temp, sizeof(temp)-1);
				temp[num_read] = '\0';
				/* Change unprintable chars */
				while (--num_read) {
					if (temp[num_read] < ' ') {
						temp[num_read] = ' ';
					}
				}
			} else {
				cout << "Timeout from tapebox" << endl;
				sleep(1);
			}
			int size = sizeof(CONSOLE_NAME)+strlen(devicename)+strlen(temp)+20;
			nameofdevice = (char *)malloc(size);
			snprintf(nameofdevice, size-1, "%s, %s, '%s'", CONSOLE_NAME, devicename, temp);
		}
		/* init device, set output and input ports */
		retval = write(fd, &(IO_INIT), sizeof(IO_INIT)-1);
		if (!retval) {
			cout << "Could not write to " << devicename << endl;
			sleep(1);
		}
		redraw();
	}

	return fd;
}

TapeboxConsole::TapeboxConsole(char *tapeboxname) {
	int i = sizeof(ondisplay);

	while (--i) {
		ondisplay[i] = 0x00;
	}
	nameofdevice = CONSOLE_NAME;
	devicename = tapeboxname;
	reconnect();
	maptexttobin();

//#ifdef DEBUG
/*
	printf("\e[2JRunning debug 2\n");
	int d = 0;
	while (d<8) {
		i = 0;
		while (i < IO_UNDEF) {
			senddata(i, 1 << d);
			i++;
		}
		printf("%d ", d); fflush(stdout);
		usleep(1000000);
		d++;
	}
	sleep(3);
	printf("\n");
*/
//#endif
}

TapeboxConsole::~TapeboxConsole(void) {

	if (fd > 0) {
		int i = 0;
		while (i < IO_UNDEF) {
			senddata(i, i==6?0x80:0 );
			i++;
		}
		//ioctl(fd, TCSETS, &term_orig);
		close(fd);
	}
	if (nameofdevice) {
		free(nameofdevice);
	}

	return;
}

int TapeboxConsole::redraw(void) {
	int i = sizeof(ondisplay);

	while (--i) {
		senddata( i, ondisplay[i]);
	}

	return sizeof(ondisplay);
}

int fdkey = open("speakreckeys.log", O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
struct timeval tv1 = {0,0};

tape_key TapeboxConsole::getcommand(fd_set *read_set, tape_key lastcmd) {
	int i, n;
	int num_read;
	struct {
		unsigned char port;
		unsigned char value;
	} buf[1 << 7];
	tape_key key = lastcmd;
	struct timeval tv2,tv3;
	int got_pressed = 0;
	int got_released = 0;
	char temp[100] = " ";

	if (FD_ISSET(fd, read_set)) {
		num_read = read(fd, &buf, sizeof(buf));
		if (num_read >= sizeof(buf)) {
			/* We got noise!! */
			snprintf(temp, sizeof(temp)-1, "Noise, num_read==%d\n", num_read);
			write(fdkey, &temp, strlen(temp));
		}
		/* Avoid noise by looking from end of buffer to start */
		//n = num_read/sizeof(buf[0]) - 1;
		n = 0;
		while (n < (num_read/2)) {
			/* if data from keyboard ... */
			if (buf[n].port == tolower(IO_INPTC)) {
				if (temp[0] == ' ') {
					gettimeofday(&tv2,NULL);
					if (tv1.tv_sec) {
						tv3.tv_sec  = tv2.tv_sec  - tv1.tv_sec;
						tv3.tv_usec = tv2.tv_usec - tv1.tv_usec;
						if (tv3.tv_usec < 0) tv3.tv_usec += 1000000;
					} else {
						tv3 = tv1;
					}
					snprintf(temp, sizeof(temp)-1, "%5d.%06d %02d: ", tv3.tv_sec % 100000, tv3.tv_usec, num_read);
					write(fdkey, &temp, strlen(temp));
					tv1 = tv2;
				}
				if (!negateinput && buf[n].value == 0xff) {
					negateinput = 0xff;
					snprintf(temp, sizeof(temp)-1, "[negateinput] ");
					write(fdkey, &temp, strlen(temp));
				}
				if (negateinput) {
					buf[n].value = ~buf[n].value;
				}
				snprintf(temp, sizeof(temp)-1, "%c=%02x ", buf[n].port, buf[n].value);
				write(fdkey, &temp, strlen(temp));
				buf[n].value &= MASK_OFF_USED_KEYS;
				if (buf[n].value) {
					if (!got_pressed) {
						got_pressed++;
						/* One or more keys pressed */
						i = 0;
						while (i < SIZE_TBKEYS && key == lastcmd) {
							if (tbkeys[i].key_code == buf[n].value) {
								key = tbkeys[i].key;
								snprintf(temp, sizeof(temp)-1, "(%d) ", key);
								write(fdkey, &temp, strlen(temp));
							}
							i++;
						}
					}
				} else {
					/* No key pressed, one must have been released */
					got_released++;
					//n = 0; /* don't look at more data in buffer, released all keys has high priority */
					//n = num_read; /* don't look at more data in buffer, released all keys has high priority */
				}
			}
			n++;
			//n--;
		}
		if (got_released) {
			/* Over rule got_pressed to avoid noise */
			if (prev_key == key_edit) {
				key = key_play;
			} else {
				key = key_none;
			}
		}
		if (got_pressed || got_pressed) {
			prev_key = key;
		}
		if (temp[0] != ' ') {
			write(fdkey, "\n", 1);
		}
	}

	return key;
}

int TapeboxConsole::fd_set_read(fd_set *read_set) {

	if (fd) {
		FD_SET(fd, read_set);
	}

	return fd;
}

int TapeboxConsole::fd_set_write(fd_set *write_set) {

	//FD_SET(fd, write_set);

	return fd;
}

unsigned int volconv[] = {
	0x0080, /* 0000 0000 1000 0000 00 */
	0x00C0, /* 0000 0000 1100 0000 01 */
	0x00E0, /* 0000 0000 1110 0000 02 */
	0x00F0, /* 0000 0000 1111 0000 03 */
	0x00F8, /* 0000 0000 1111 1000 04 */
	0x00FC, /* 0000 0000 1111 1100 05 */
	0x00FE, /* 0000 0000 1111 1110 06 */
	0x00FF, /* 0000 0000 1111 1111 07 */
	0x80FF, /* 1000 0000 1111 1111 08 */
	0xC0FF, /* 1100 0000 1111 1111 09 */
	0xE0FF, /* 1110 0000 1111 1111 10 */
	0xF0FF, /* 1111 0000 1111 1111 11 */
	0xF8FF, /* 1111 1000 1111 1111 12 */
	0xFCFF, /* 1111 1100 1111 1111 13 */
	0xFEFF, /* 1111 1110 1111 1111 14 */
	0xFFFF  /* 1111 1111 1111 1111 15 */
};
#define SIZE_VOLCONV (sizeof(volconv) / sizeof(volconv[0]))

unsigned int peakconv[] = {
	0x0080, /* 0000 0000 1000 0000 00 */
	0x0040, /* 0000 0000 0100 0000 01 */
	0x0020, /* 0000 0000 0010 0000 02 */
	0x0010, /* 0000 0000 0001 0000 03 */
	0x0008, /* 0000 0000 0000 1000 04 */
	0x0004, /* 0000 0000 0000 0100 05 */
	0x0002, /* 0000 0000 0000 0010 06 */
	0x0001, /* 0000 0000 0000 0001 07 */
	0x8000, /* 1000 0000 0000 0000 08 */
	0x4000, /* 0100 0000 0000 0000 09 */
	0x2000, /* 0010 0000 0000 0000 10 */
	0x1000, /* 0001 0000 0000 0000 11 */
	0x0800, /* 0000 1000 0000 0000 12 */
	0x0400, /* 0000 0100 0000 0000 13 */
	0x0200, /* 0000 0010 0000 0000 14 */
	0x0100  /* 0000 0001 0000 0000 15 */
};
#define SIZE_PEAKCONV (sizeof(peakconv) / sizeof(peakconv[0]))

int TapeboxConsole::setvolume(int newvol, int peakvol) {
	unsigned int binvol;

	binvol = volconv[(newvol / 4) % SIZE_VOLCONV] | peakconv[(peakvol / 4) % SIZE_PEAKCONV];
	if ((binvol & 0xFF) != (lastbinvol & 0xFF)) {
		senddata(IO_VOLUME_LOW,  (binvol >> 0) & 0xFF);
	}
	if ((binvol & 0xFF00) != (lastbinvol & 0xFF00)) {
		senddata(IO_VOLUME_HIGH, (binvol >> 8) & 0xFF);
	}
	lastbinvol = binvol;

	return 0;
}

int TapeboxConsole::showsegment( int num, unsigned char value ) {
	unsigned char bin;

	value = toupper(value);
	bin = segmentbin[value & 0x7F] | value & 0x80; /* dp */
	senddata(IO_SEGMENT+num, bin);

	return bin;
}

#define SIZE_BOXBUTTON (sizeof(boxbutton) / sizeof(boxbutton[0]))

int TapeboxConsole::setbuttonon(button_name name) {
	int i = 0;
	int bin = 0;

	while (i < SIZE_TBKEYS) {
		if (tbkeys[i].but_name == name) {
			bin = tbkeys[i].led_code;
		}
		i++;
	}
	senddata(IO_LED, bin );

	return bin;
}

int TapeboxConsole::help(int outdev) {

	dprintf(outdev, "Help for tapebox.\n");
	dprintf(outdev, "+---+---+---+---+\n");
	dprintf(outdev, "| P | M | V | T |\n");
	dprintf(outdev, "+---+---+---+---+\n");
	dprintf(outdev, "+---+ +---+ +---+\n");
	dprintf(outdev, "| R | | E | | S |\n");
	dprintf(outdev, "+---+ +---+ +---+\n");

	return 0;
}

